//: Render a tessellated quad patch
/*
 
 Instead of rendering with vertices, we render a patch with four control points.
 Instead of setting up just a normal render pipeline state,
 we set up a compute pipeline state to do the tessellation
 and a render pipeline state with tessellation properties
 to evaluate the tessellation (move the vertices where we want them)
 
 The four control points are defined in Points.swift
 One of the points has z=1.0 and the other three are z=0.
 This will raise one of the corners and all vertices
 in between the other control points are interpolated
 using the bilinear interpolation formula
 This forms a curved surface (more curved with
 more tessellation)
*/
 


import Cocoa
import MetalKit
import PlaygroundSupport

// Setup Metal
let device = MTLCreateSystemDefaultDevice()!
let queue = device.makeCommandQueue()!
let frame = NSRect(x: 0, y: 0, width: 600, height: 600)
let metalView = MetalView(frame: frame, device: device)

let renderer = Renderer(device: device)
// change these to up the tessellation quality
//renderer.edgeFactor = [48]
//renderer.insideFactor = [48]

metalView.delegate = renderer

PlaygroundPage.current.liveView = metalView


