#include <metal_stdlib>
using namespace metal;

struct ModelConstants {
  float4x4 modelViewMatrix;
};

// Control Point struct
struct ControlPoint {
  float4 position [[attribute(0)]];
};

// Patch struct
struct PatchIn {
  patch_control_point<ControlPoint> control_points;
};

// Vertex-to-Fragment struct
struct VertexOut {
  float4 position [[position]];
  half4  color [[flat]];
};

// Quad compute kernel
kernel void tessellation_kernel_quad(constant float& edge_factor [[ buffer(0) ]],
                                     constant float& inside_factor [[ buffer(1) ]],
                                     device MTLQuadTessellationFactorsHalf* factors [[ buffer(2) ]],
                                     uint pid [[ thread_position_in_grid ]])
{
  factors[pid].edgeTessellationFactor[0] = edge_factor;
  factors[pid].edgeTessellationFactor[1] = edge_factor;
  factors[pid].edgeTessellationFactor[2] = edge_factor;
  factors[pid].edgeTessellationFactor[3] = edge_factor;
  factors[pid].insideTessellationFactor[0] = inside_factor;
  factors[pid].insideTessellationFactor[1] = inside_factor;
}

fragment half4 tessellation_fragment(VertexOut fragmentIn [[stage_in]])
{
  return fragmentIn.color;
}


// Quad post-tessellation vertex function
// Bilinear patch with four control points
[[patch(quad, 4)]]
vertex VertexOut tessellation_sphere(PatchIn patchIn [[stage_in]],
                                          float2 patch_coord [[ position_in_patch ]],
                                          constant ModelConstants &modelConstants [[ buffer(1) ]])
{
  VertexOut vertexOut;

  float u = patch_coord.x;
  float v = patch_coord.y;

  float r = 0.2;
  float pi2 = 6.28318530*2;
  float pi = pi2*0.5;
  
  float phi = pi * u;
  float theta = pi2 * v;
  float sinPhi, cosPhi, sinTheta, cosTheta;
  sinPhi = sin(phi);
  cosPhi = cos(phi);
  sinTheta = sin(theta);
  cosTheta = cos(theta);
  
  float3 torus;
  torus.x = (0.5+r*cosTheta)*cosPhi;
  torus.y = (0.5+r*cosTheta)*sinPhi;
  torus.z = r * sinTheta;
  
  
  vertexOut.position = float4(torus, 1);
  vertexOut.position = modelConstants.modelViewMatrix * vertexOut.position;
  torus = float3(normalize(torus) + 0.4);
  vertexOut.color = half4(half3(torus), 1);
  return vertexOut;
}


