#include <metal_stdlib>
using namespace metal;

struct ModelConstants {
  float4x4 modelViewMatrix;
};

// Control Point struct
struct ControlPoint {
  float4 position [[attribute(0)]];
};

// Patch struct
struct PatchIn {
  patch_control_point<ControlPoint> control_points;
};

// Vertex-to-Fragment struct
struct VertexOut {
  float4 position [[position]];
  half4  color [[flat]];
};

// Quad compute kernel
kernel void tessellation_kernel_quad(constant float& edge_factor [[ buffer(0) ]],
                                     constant float& inside_factor [[ buffer(1) ]],
                                     device MTLQuadTessellationFactorsHalf* factors [[ buffer(2) ]],
                                     uint pid [[ thread_position_in_grid ]])
{
  factors[pid].edgeTessellationFactor[0] = edge_factor;
  factors[pid].edgeTessellationFactor[1] = edge_factor;
  factors[pid].edgeTessellationFactor[2] = edge_factor;
  factors[pid].edgeTessellationFactor[3] = edge_factor;
  factors[pid].insideTessellationFactor[0] = inside_factor;
  factors[pid].insideTessellationFactor[1] = inside_factor;
}

fragment half4 tessellation_fragment(VertexOut fragmentIn [[stage_in]])
{
  return fragmentIn.color;
}

[[patch(quad, 4)]]
vertex VertexOut tessellation_horn(PatchIn patchIn [[stage_in]],
                                          float2 patch_coord [[ position_in_patch ]],
                                          constant ModelConstants &modelConstants [[ buffer(1) ]])
{
  VertexOut vertexOut;

  float u = patch_coord.x;
  float v = patch_coord.y;

  float pi2 = 6.28318530;
  float pi = pi2*0.5;
  
  float phi = 1 * u;
  float theta = pi2 * v;
  float sinPhi, cosPhi, sinTheta, cosTheta;
  sinPhi = sin(phi);
  cosPhi = cos(phi);
  sinTheta = sin(theta);
  cosTheta = cos(theta);
  
  //u is phi and v is theta
  float3 shape;
  float a = 2;
  shape.x = (a+phi*cosTheta)*sin(a*pi*phi);
  shape.y = (a+phi*cosTheta)*cos(a*pi*phi)+a*phi;
  shape.z = phi * sinTheta;
  
  vertexOut.position = float4(shape, 1);
  vertexOut.position = modelConstants.modelViewMatrix * vertexOut.position;
  float3 colorShape = float3(normalize(shape) + 0.4);
  vertexOut.color = half4(half3(colorShape), 1);
  return vertexOut;
}


[[patch(quad, 4)]]
vertex VertexOut tessellation_apple(PatchIn patchIn [[stage_in]],
                                   float2 patch_coord [[ position_in_patch ]],
                                   constant ModelConstants &modelConstants [[ buffer(1) ]])
{
  VertexOut vertexOut;
  
  float u = patch_coord.x;
  float v = patch_coord.y;
  
  float pi2 = 6.28318530;
  float pi = pi2*0.5;
  
  float fi = pi2 * u;
  float theta = pi * v;
  float sinFi, cosFi, sinTheta, cosTheta;
  sinFi = sin(fi);
  cosFi = cos(fi);
  sinTheta = sin(theta);
  cosTheta = cos(theta);
  
  //u is fi and v is theta
  float3 shape;
  shape.x = cosFi*(4+3.8*cosTheta);
  shape.y = sinFi*(4+3.8*cosTheta);
  shape.z = (cosTheta+sinTheta-1)*(1+sinTheta)*log(1-pi*theta/10)+7.5*sinTheta;
  
  vertexOut.position = float4(shape, 1);
  vertexOut.position = modelConstants.modelViewMatrix * vertexOut.position;
  float3 colorShape = float3(normalize(shape) + 0.4);
  vertexOut.color = half4(half3(colorShape), 1);
  return vertexOut;
}
